--[[ File
@description:
    This file contains source code for mapping action to output update with respect time.Q-value calculated 
    using Q-learning function. In order to collect a set amount of history (experiances) the model uses a                                                                                          		  Epsilon Greedy strategy to select at random an action. The resulting state, reward and action are                                                                                              			recorded in within the replay memory of the Agent. Where a memory refers to an array. This array maintains                                                                                      		a set size - new items replace the oldest values. It is expected that this process would be beneficial to 
the model for it to improve more quickly over time, that is if the model is capable of learning.CNN neural network used to model       
the given Q-network. 

@author: Yashashree Jadhav   email:yashashreejadhav88@gmail.com
@acknowledgement:
    RedOne Technology Pvt. Ltd. 
    Sunchon National University (IE-LAB)
]]


require 'torch'
require 'TransitionTable'
--if not dqn then
--    require 'initenv'
--end

--local classic = require 'classic'

print('loaded Agent.lua')

--local Agent = classic.torch('Agent')

local Agent = torch.class('Agent')

--[[function Agent:init(create agent and initailize. set of environment states S ;
  set of actions A. Handel transitioning between states; determine the                                                                                                                    	              scalar immediate reward of a transition)

end--]]

function Agent:_init(opt)
  print('Agent:_init()')
  
  self.height = opt.height
  self.width = opt.width
  self.nActions = opt.nActions

  --Epsion Annealing
  self.epStart = 1 --Exploration probability
  self.ep= 1 --Starting epsilon value
  self.epEnd = 0
  self.epT = 10000
  self.epEndt = 10000

  --Learning rate Annealing
  self.lrStart = .01 --Learning rate.
  self.lrEnd = 0
  self.lrEndT = 10000
  self.minibatchSize = 1
  self.validSize = 500

  --Q params
  self.gamma = .99 --Discount factor.
  self.updateFreq = 1
  -- Number of points to replay per learning step.
  self.nReplay = 1
  -- Number of steps after which learning starts.
  self.learnStart = 0

  --Transition table
  self.replayMemory = 1000
  self.histLen = 4
  self.maxReward = 5
  self.minReward = 0
  self.bestQ = 0

  self.nChannels = 1 -- number of color channels in input
  self.inputDims = {self.histLen * self.nChannels, self.height, self.width}
  self.histSpacing = 1
  self.nonTermProb = 1
  self.bufferSize = 206

 
  --self.transitionParams = {}
  print('creating Agent network... ')
  self.network = self:createNetwork()

  print('created Agent network... ')
  print(self.network)
  
  self.network:float()
  self.tensor_type = torch.FloatTensor

  local transitionArgs = 
  {
        stateDim = self.width * self.height, numActions = self.nActions,
        histLen = self.histLen, gpu = 0,
        maxSize = self.replayMemory, histType = self.histType,
        histSpacing = self.histSpacing, nonTermProb = self.nonTermProb,
        bufferSize = self.bufferSize
   }

 -- Create transition table.
    ---- assuming the transition table always gets floating point input
    ---- (Foat or Cuda tensors) and always returns one of the two, as required
    ---- internally it always uses ByteTensors for states, scaling and
    ---- converting accordingly

	--local trans = require 'TransitionTable'
    self.transitions = TransitionTable{transitionArgs}
	print('transition table')
	print(trans)
	print(self.transitions:GetNumActions())
	print(self.transitions:size())
	
    self.numSteps = 0  -- Number of perceived states.
    self.lastState = nil
    self.lastAction = nil
    self.vAvg = 0      -- V running average.
    self.tdErrAvg = 0  -- TD error running average.

    self.qMax = 1
    self.rMax = 1

    self.theta, self.dTheta = self.network:getParameters()
    self.dTheta:zero()

    self.deltas = self.dTheta:clone():fill(0)

    self.tmp = self.dTheta:clone():fill(0)
    self.g = self.dTheta:clone():fill(0)
    self.g2 = self.dTheta:clone():fill(0)

	print('done creating Agent')

end

--[[function Agent:update( agent upddated with current situation.describe what the agent observes.)

end--]]

function Agent:update(args)
  local s, a, r, s2, term, delta
  local q, q2, q2Max

  s = args.s
  a = args.a
  r = args.r
  s2 = args.s2
  term = args.term

   -- The order of calls to forward is a bit odd in order
    -- to avoid unnecessary calls (we only need 2).


  -- delta = r + (1-terminal) * gamma * max_a Q(s2, a) - Q(s, a)
  term = term:clone():float():mul(-1):add(1)

  targetQ = self.network

  --Comput max_a Q(s_2,a)
  q2Max = targetQ:forward(s2):float():max(2)
  -- Compute q2 = (1-terminal) * gamma * max_a Q(s2, a)
  q2 = q2Max:clone():mul(self.discount):cmul(term)
  delta = r:clone():float()
  delta:add(q2)

   -- q = Q(s,a)
   local qAll = self.network:forward(s):float()
   q = torch.FloatTensor(qAll:size(1))
   for i=1,qAll:size(1) do
       q[i] = qAll[i][a[i]]
   end
   delta:add(-1, q)

   local targets = torch.zeros(self.minibatchSize, self.nActions):float()
   for i=1,math.min(self.minibatchSize,a:size(1)) do
      targets[i][a[i]] = delta[i]
   end

   return targets, delta, q2Max
 end


--[[function Agent:learn( agent learning as per algorithm)

end--]]

function Agent:learn()
  --Perform a minibatch update
  -- Perform a minibatch Q-learning update:
    -- w += alpha * (r + gamma max Q(s2,a2) - Q(s,a)) * dQ(s,a)/dw
	print('Agent:learn()')
	
  local s, a, r, s2, term = self.transitions:sample(self.minibatchSize)

  print('  calling Agent:update()')
	
  local targets, delta, q2Max = self:update{s=s, a=a, r=r, s2=s2, term=term, updateQmax=true}

  print('  calling self.dTheta:zero()')
	
  --Zero grab params
  self.dTheta:zero()
  
   print('  calling self.network:backwards()')
  --Compute new gradient
  self.network:backwards(s, targets)

  print('  done self.network:backwards()')
  
  --Anneal Learning rate
  local t = math.max(0, self.numSteps - self.learnStart)
  self.lr = (self.lrStart - self.lrEnd) * (self.lrEndT - t)/(self.lrEndT + self.lrEnd)
  self.lr = math.max(self.lr, self.lrEnd)

  --Compute RMSprop (gradients)
  self.g:mul(0.95):add(0.05, self.dw)
  self.tmp:cmul(self.dw, self.dw)
  self.g2:mul(0.95):add(0.05, self.tmp)
  self.tmp:cmul(self.g, self.g)
  self.tmp:mul(-1)
  self.tmp:add(self.g2)
  self.tmp:add(0.01)
  self.tmp:sqrt()

  --Accumulate update
  self.deltas:mul(0):addcdiv(self.lr, self.dTheta, self.tmp)
  self.theta:add(self.deltas)
end


--[[function Agent:perceive(agent perceive rewards, state, terminal information. Call Agent:perceive function
    from mobilerobot.lua file )

end--]]

function Agent:perceive(reward, state, terminal, testing, testingEp)
  local curState
  if self.maxReward then
    reward = math.min(reward, self.maxReward)
  end
  if self.minReward then
    reward = math.max(reward, self.minReward)
  end

   print(reward)
   print(terminal) 
   -- print(state)
   image.save('AgentReciveImage.png',state)

--	print('Transition Table:  ' .. self.transitions:size())
	self.transitions:addRecentState(state, terminal)
	local currentFullState = self.transitions:getRecent()

  --store transitions  s, a, r, s'
  if self.lastState and not testing then
	print('self.transitions:add(self.lastState, self.lastAction')
    self.transitions:add(self.lastState, self.lastAction, reward, self.lastTerminal, priority)
  end


      if self.numSteps == self.learn_start+1 and not testing then
        self:sample_validation_data()
    end
 -- print('getcurState')
--  curState = self.transitions:getRecent()
--  print('curState')
--  print(curState:size())
--  curState = curState:resize(1, unpack(self.inputDims))

      -- Select action
  local actionIndex = 1
  if not terminal then
    actionIndex = self.eGreedy(curState, testingEp)	
  end

  self.transitions:addRecentAction(actionIndex)
  
  print('update Q Learner')
  
  --Update q learner (Do some Q-learning updates)
  if self.numSteps > self.learnStart and not testing and self.numSteps % self.updateFreq == 0 then
    for i = 1, self.nReplay do
       self:learn()
     end
   end

   if not testing then
     self.numSteps = self.numSteps + 1
   end

   self.lastState = state:clone()
   self.lastAction = actionIndex
   self.lastTerminal = terminal

   if self.target_q and self.numSteps % self.target_q == 1 then
    self.target_network = self.network:clone()
  end




  if not terminal then
    return actionIndex
  else
    return 0
  end
end


--[[function Agent:eGreedy(the agent chooses the action that it believes has the best long-term effect with                                                                                            	probability 1 − ϵ , and it chooses an action uniformly at random )
end--]]

function Agent:eGreedy(state, testingEp)
   print('Agent:eGreedy()')
  self.ep = testingEp or (self.epEnd +
                          math.max(0, (self.epStart - self.epEnd) * (self.epEndt -
                          math.max(0, self.numSteps - self.learnStart))/self.epEndt))
    -- Epsilon greedy
          if torch.uniform() < self.ep then
			print('eGreedy -- randomizing')
            return torch.random(1, self.nActions)
          else
			print('eGreedy -- greedy network:forward')
            return self:greedy(state)
          end
end

function Agent:greedy(state)
-- Turn single state into minibatch.  Needed for convolutional nets.
  if state:dim() == 2 then
    assert(false, 'input must be at least 3d')
    state = state:resize(1, state:size(1), state:size(2))
  end

  local q = self.network:forward(state):float():squeeze()
  local maxq = q[1]
  local besta = {1}


    -- Evaluate all other actions (with random tie-breaking)
    for a = 2, self.n_actions do
        if q[a] > maxq then
            besta = { a }
            maxq = q[a]
        elseif q[a] == maxq then
            besta[#besta+1] = a
        end
    end
    self.bestQ = maxq

    local r = torch.random(1, #besta)

    self.lastAction = besta[4]

    return besta[r]
end





--[[function Agent:createNetwork()( Create CNN network here )
end--]]

function Agent:createNetwork()
  local hiddenSize = 128
  local model = nn.Sequential()
  model:add(nn.View(self.histLen, self.height, self.width))
  model:add(nn.SpatialConvolution(self.histLen*self.nChannels, 32, 5, 5, 2, 2, 1, 1))
  model:add(nn.ReLU(true))
  model:add(nn.SpatialConvolution(32,32,4,4,2,2))
  model:add(nn.ReLU(true))
  local convOutputSize = torch.prod(torch.Tensor(model:forward(torch.Tensor(torch.LongStorage({self.histLen*self.nChannels, self.height, self.width}))):size():totable()))
  print('convOutputSize:  ' .. convOutputSize)
  model:add(nn.Linear(convOutputSize, hiddenSize))
  model:add(nn.ReLU(true))
  model:add(nn.Linear(hiddenSize, self.nActions))
  return model
end



function Agent:init(arg)
     self.nActions = opt.nActions
    self.network = self:_loadNet()
    -- Generate targets.
    self.transitions:empty()

end


function Agent:report()
    print(get_weight_norms(self.network))
    print(get_grad_norms(self.network))
end



return Agent

